///////////////////////////////////////////////////////////////////////////////////////////////////
//
// PROJECT : Arirang family
//
// DATE    : 2008.10.23
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FILE  : Arirang_4S8.h
//
// NOTES : Optimized code of Arirang family
// 
//         Based on 8-bit platform (4 8-bit S-box version)
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdlib.h>


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Macro

// Define the Endianness
#undef BIG_ENDIAN
#undef LITTLE_ENDIAN

#if defined(USER_BIG_ENDIAN)
	#define BIG_ENDIAN
#elif defined(USER_LITTLE_ENDIAN)
	#define LITTLE_ENDIAN
#else
	#if 0
		#define BIG_ENDIAN
	#elif defined(_MSC_VER)
		#define LITTLE_ENDIAN
	#else
		#error
	#endif
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Definition

#define ARIRANG224_BLOCK_LEN	64
#define ARIRANG224_DIGEST_LEN	28

#define ARIRANG256_BLOCK_LEN	64
#define ARIRANG256_DIGEST_LEN	32

#define ARIRANG384_BLOCK_LEN	128
#define ARIRANG384_DIGEST_LEN	48

#define ARIRANG512_BLOCK_LEN	128
#define ARIRANG512_DIGEST_LEN	64


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Datatype

typedef unsigned char BYTE;				//  8-bit data type

typedef unsigned char BitSequence;		//  8-bit data type


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Error

typedef enum {SUCCESS = 0, FAIL = 1, BAD_HASHLEN = 2} HashReturn;


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Structure

typedef struct {

	// Hash length
	BYTE hashbytelen;
	
	// Counter
	BYTE counter[16];

	// Count
	BYTE count[16];

	// Message block
	BYTE block[ARIRANG512_BLOCK_LEN];

	// Working variables
	BYTE workingvar[ARIRANG512_DIGEST_LEN];

	// hash block length
	BYTE blocklen;

	// bit_length % 7
	BYTE remainderbit;


} hashState;


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Function

void gen_tabs(void);

HashReturn Init(hashState *state, BYTE hashbytelen);

HashReturn Update(hashState *state, const BitSequence *data, BYTE *databitlen);

HashReturn Final(hashState *state, BYTE *hashval);

HashReturn Hash(BYTE hashbytelen, const BitSequence *data, BYTE *databitlen, BitSequence *hashval);
