///////////////////////////////////////////////////////////////////////////////////////////////////
//
// PROJECT : ARIRANG family(ARIRANG-224,ARIRANG-256,ARIRANG-384,ARIRANG-512)
//
// DATE    : 2008.10.23
//
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FILE  : ARIRANG_1S8.c
//
// NOTES : Code of ARIRANG family 
// 
//         Based on 8-bit platform (one s-box version)
//
///////////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>

#include <memory.h>

#include "Arirang_1S8.h"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Macro

#undef LL
#undef RR
#define LL(x,n)		(x << n)
#define RR(x,n)		(x >> n)

#undef SS0
#undef SS1
#undef SS2
#undef SS3
#undef SSS0
#undef SSS1
#undef SSS2
#undef SSS3
#undef SSS4
#undef SSS5
#undef SSS6
#undef SSS7

#define SS0(x,y,n1,n2)	x[0] ^= RR(y[1],n1)^LL(y[0],n2); x[1] ^= RR(y[2],n1)^LL(y[1],n2); x[2] ^= RR(y[3],n1)^LL(y[2],n2); x[3] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS1(x,y,n1,n2)	x[3] ^= RR(y[1],n1)^LL(y[0],n2); x[0] ^= RR(y[2],n1)^LL(y[1],n2); x[1] ^= RR(y[3],n1)^LL(y[2],n2); x[2] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS2(x,y,n1,n2)	x[2] ^= RR(y[1],n1)^LL(y[0],n2); x[3] ^= RR(y[2],n1)^LL(y[1],n2); x[0] ^= RR(y[3],n1)^LL(y[2],n2); x[1] ^= RR(y[0],n1)^LL(y[3],n2);
#define SS3(x,y,n1,n2)	x[1] ^= RR(y[1],n1)^LL(y[0],n2); x[2] ^= RR(y[2],n1)^LL(y[1],n2); x[3] ^= RR(y[3],n1)^LL(y[2],n2); x[0] ^= RR(y[0],n1)^LL(y[3],n2);

#define SSS0(x,y,n1,n2)	x[0] ^= RR(y[1],n1)^LL(y[0],n2); x[1] ^= RR(y[2],n1)^LL(y[1],n2); x[2] ^= RR(y[3],n1)^LL(y[2],n2); x[3] ^= RR(y[4],n1)^LL(y[3],n2); x[4] ^= RR(y[5],n1)^LL(y[4],n2); x[5] ^= RR(y[6],n1)^LL(y[5],n2); x[6] ^= RR(y[7],n1)^LL(y[6],n2); x[7] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS1(x,y,n1,n2)	x[7] ^= RR(y[1],n1)^LL(y[0],n2); x[0] ^= RR(y[2],n1)^LL(y[1],n2); x[1] ^= RR(y[3],n1)^LL(y[2],n2); x[2] ^= RR(y[4],n1)^LL(y[3],n2); x[3] ^= RR(y[5],n1)^LL(y[4],n2); x[4] ^= RR(y[6],n1)^LL(y[5],n2); x[5] ^= RR(y[7],n1)^LL(y[6],n2); x[6] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS2(x,y,n1,n2)	x[6] ^= RR(y[1],n1)^LL(y[0],n2); x[7] ^= RR(y[2],n1)^LL(y[1],n2); x[0] ^= RR(y[3],n1)^LL(y[2],n2); x[1] ^= RR(y[4],n1)^LL(y[3],n2); x[2] ^= RR(y[5],n1)^LL(y[4],n2); x[3] ^= RR(y[6],n1)^LL(y[5],n2); x[4] ^= RR(y[7],n1)^LL(y[6],n2); x[5] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS3(x,y,n1,n2)	x[5] ^= RR(y[1],n1)^LL(y[0],n2); x[6] ^= RR(y[2],n1)^LL(y[1],n2); x[7] ^= RR(y[3],n1)^LL(y[2],n2); x[0] ^= RR(y[4],n1)^LL(y[3],n2); x[1] ^= RR(y[5],n1)^LL(y[4],n2); x[2] ^= RR(y[6],n1)^LL(y[5],n2); x[3] ^= RR(y[7],n1)^LL(y[6],n2); x[4] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS4(x,y,n1,n2)	x[4] ^= RR(y[1],n1)^LL(y[0],n2); x[5] ^= RR(y[2],n1)^LL(y[1],n2); x[6] ^= RR(y[3],n1)^LL(y[2],n2); x[7] ^= RR(y[4],n1)^LL(y[3],n2); x[0] ^= RR(y[5],n1)^LL(y[4],n2); x[1] ^= RR(y[6],n1)^LL(y[5],n2); x[2] ^= RR(y[7],n1)^LL(y[6],n2); x[3] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS5(x,y,n1,n2)	x[3] ^= RR(y[1],n1)^LL(y[0],n2); x[4] ^= RR(y[2],n1)^LL(y[1],n2); x[5] ^= RR(y[3],n1)^LL(y[2],n2); x[6] ^= RR(y[4],n1)^LL(y[3],n2); x[7] ^= RR(y[5],n1)^LL(y[4],n2); x[0] ^= RR(y[6],n1)^LL(y[5],n2); x[1] ^= RR(y[7],n1)^LL(y[6],n2); x[2] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS6(x,y,n1,n2)	x[2] ^= RR(y[1],n1)^LL(y[0],n2); x[3] ^= RR(y[2],n1)^LL(y[1],n2); x[4] ^= RR(y[3],n1)^LL(y[2],n2); x[5] ^= RR(y[4],n1)^LL(y[3],n2); x[6] ^= RR(y[5],n1)^LL(y[4],n2); x[7] ^= RR(y[6],n1)^LL(y[5],n2); x[0] ^= RR(y[7],n1)^LL(y[6],n2); x[1] ^= RR(y[0],n1)^LL(y[7],n2);
#define SSS7(x,y,n1,n2)	x[1] ^= RR(y[1],n1)^LL(y[0],n2); x[2] ^= RR(y[2],n1)^LL(y[1],n2); x[3] ^= RR(y[3],n1)^LL(y[2],n2); x[4] ^= RR(y[4],n1)^LL(y[3],n2); x[5] ^= RR(y[5],n1)^LL(y[4],n2); x[6] ^= RR(y[6],n1)^LL(y[5],n2); x[7] ^= RR(y[7],n1)^LL(y[6],n2); x[0] ^= RR(y[0],n1)^LL(y[7],n2);

#define f_mult2(x) (x&0x80 ? (0x1b^(x<<1)) : (x<<1))
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// Constant
static const BYTE K256[64] = {
	0x51, 0x7c, 0xc1, 0xb7, 0x76, 0x51, 0x7c, 0xc1, 0xbd, 0x76, 0x51, 0x7c, 0x2d, 0xbd, 0x76, 0x51,
	0x27, 0x2d, 0xbd, 0x76, 0xcb, 0x27, 0x2d, 0xbd, 0x90, 0xcb, 0x27, 0x2d, 0x0a, 0x90, 0xcb, 0x27,
	0xec, 0x0a, 0x90, 0xcb, 0x5b, 0xec, 0x0a, 0x90, 0x9a, 0x5b, 0xec, 0x0a, 0xe6, 0x9a, 0x5b, 0xec,
	0xb7, 0xe6, 0x9a, 0x5b, 0xc1, 0xb7, 0xe6, 0x9a, 0x7c, 0xc1, 0xb7, 0xe6, 0x51, 0x7c, 0xc1, 0xb7
};

static const BYTE K512[128] = {
	0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94, 0x2d, 0xb6, 0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0xe6, 0x95, 0x2d, 0xb6, 0x51, 0x7c, 0xc1, 0xb7, 0x90, 0xcb, 0xe6, 0x95, 0x2d, 0xb6, 0x51, 0x7c,
	0x7c, 0xca, 0x90, 0xcb, 0xe6, 0x95, 0x2d, 0xb6, 0xcb, 0x23, 0x7c, 0xca, 0x90, 0xcb, 0xe6, 0x95, 0x76, 0x5e, 0xcb, 0x23, 0x7c, 0xca, 0x90, 0xcb, 0xec, 0x01, 0x76, 0x5e, 0xcb, 0x23, 0x7c, 0xca,
	0xb7, 0xe9, 0xec, 0x01, 0x76, 0x5e, 0xcb, 0x23, 0xbd, 0x7d, 0xb7, 0xe9, 0xec, 0x01, 0x76, 0x5e, 0x9a, 0x5f, 0xbd, 0x7d, 0xb7, 0xe9, 0xec, 0x01, 0x5b, 0xe8, 0x9a, 0x5f, 0xbd, 0x7d, 0xb7, 0xe9,
	0x0a, 0x94, 0x5b, 0xe8, 0x9a, 0x5f, 0xbd, 0x7d, 0x27, 0x22, 0x0a, 0x94, 0x5b, 0xe8, 0x9a, 0x5f, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94, 0x5b, 0xe8, 0x51, 0x7c, 0xc1, 0xb7, 0x27, 0x22, 0x0a, 0x94
};

BYTE	sbx1[256]; // Prepare S-box: 

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : gen_tabs
//
// DESCRIPTION    : Generate the s-box table and field multiplication table
//
// PARAMETERS     : void
//                  
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void gen_tabs(void)
{   
	BYTE i; 	
    BYTE p, q;
	BYTE pow_tab[256],log_tab[256];
	
 
	/* log and power tables for GF(2**8) finite field with  */
    /* 0x011b as modular polynomial - the simplest prmitive */
    /* root is 0x03, used here to generate the tables       */

	i = 0;p = 1;
    while(1)
    {
        pow_tab[i] = (BYTE)p;
		log_tab[p] = (BYTE)i;  
        p = p ^ (p << 1) ^ (p & 0x80 ? 0x01b : 0);
		if(i == 255) break;
		i++;
    }

    /* note that the affine byte transformation matrix in   */
    /* ARIRANG specification is in big endian format with  */
    /* bit 0 as the most significant bit. In the remainder  */
    /* of the specification the bits are numbered from the  */
    /* least significant end of a byte.                     */

	i = 0;
	while(1)    
    {   
        p = (i ? pow_tab[255 - log_tab[i]] : 0); q = p; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q; 
        q = (q >> 7) | (q << 1); p ^= q ^ 0x63; 
        sbx1[i] = p;
		if(i == 0xff) break;
		i++;
    }	
}
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step256
//
// DESCRIPTION    : Step function of ARIRANG-224 and ARIRANG-256
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Step256(BYTE* A, BYTE* B, BYTE* C, BYTE* D, BYTE* E, BYTE* F, BYTE* G, BYTE* H, BYTE* M1, BYTE* M2){
	BYTE temp1[4], ff[2][4];
	A[0] ^= M1[0]; A[1] ^= M1[1]; A[2] ^= M1[2]; A[3] ^= M1[3];							
	E[0] ^= M2[0]; E[1] ^= M2[1]; E[2] ^= M2[2]; E[3] ^= M2[3];							
	ff[0][3]=sbx1[A[3]];	ff[1][3]=f_mult2(ff[0][3]);
	ff[0][2]=sbx1[A[2]];	ff[1][2]=f_mult2(ff[0][2]);
	ff[0][1]=sbx1[A[1]];	ff[1][1]=f_mult2(ff[0][1]);
	ff[0][0]=sbx1[A[0]];	ff[1][0]=f_mult2(ff[0][0]);
	temp1[3] = ff[1][3] ^(ff[1][2] ^ ff[0][2])^ ff[0][1] ^ ff[0][0];
	temp1[2] = ff[0][3] ^ ff[1][2] ^(ff[1][1] ^ ff[0][1])^ ff[0][0];
	temp1[1] = ff[0][3] ^ ff[0][2] ^ ff[1][1] ^(ff[1][0] ^ ff[0][0]);
	temp1[0] =(ff[1][3] ^ ff[0][3])^ ff[0][2] ^ ff[0][1] ^ ff[1][0];
	B[0] ^= temp1[0];	B[1] ^= temp1[1];	B[2] ^= temp1[2];	B[3] ^= temp1[3];		
	SS1(C,temp1,3,5);																	
	SS2(D,temp1,1,7);																	
	ff[0][3]=sbx1[E[3]];	ff[1][3]=f_mult2(ff[0][3]);
	ff[0][2]=sbx1[E[2]];	ff[1][2]=f_mult2(ff[0][2]);
	ff[0][1]=sbx1[E[1]];	ff[1][1]=f_mult2(ff[0][1]);
	ff[0][0]=sbx1[E[0]];	ff[1][0]=f_mult2(ff[0][0]);
	temp1[3] = ff[1][3] ^(ff[1][2] ^ ff[0][2])^ ff[0][1] ^ ff[0][0];
	temp1[2] = ff[0][3] ^ ff[1][2] ^(ff[1][1] ^ ff[0][1])^ ff[0][0];
	temp1[1] = ff[0][3] ^ ff[0][2] ^ ff[1][1] ^(ff[1][0] ^ ff[0][0]);
	temp1[0] =(ff[1][3] ^ ff[0][3])^ ff[0][2] ^ ff[0][1] ^ ff[1][0];					
	F[0] ^= temp1[0];	F[1] ^= temp1[1];	F[2] ^= temp1[2];	F[3] ^= temp1[3];		
	SS3(G,temp1,3,5);																	
	SS0(H,temp1,1,7);																	
}																						
																						
///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : step512
//
// DESCRIPTION    : Step funCompression function of ARIRANG-384 and Arirang-512
//
// PARAMETERS     : A,B,C,D,E,F,G,H  - working variables
//                  M1,M2 - message block
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Step512(BYTE* A, BYTE* B, BYTE* C, BYTE* D, BYTE* E, BYTE* F, BYTE* G, BYTE* H, BYTE* M1, BYTE* M2){																		
	BYTE temp1[8], ff[4][8];
	A[0] ^= M1[0]; A[1] ^= M1[1]; A[2] ^= M1[2]; A[3] ^= M1[3];															
	A[4] ^= M1[4]; A[5] ^= M1[5]; A[6] ^= M1[6]; A[7] ^= M1[7];															
	E[0] ^= M2[0]; E[1] ^= M2[1]; E[2] ^= M2[2]; E[3] ^= M2[3];															
	E[4] ^= M2[4]; E[5] ^= M2[5]; E[6] ^= M2[6]; E[7] ^= M2[7];															
	ff[0][7]=sbx1[A[7]]; ff[1][7]=f_mult2(ff[0][7]); ff[2][7]=f_mult2(ff[1][7]); ff[3][7]=f_mult2(ff[2][7]);
	ff[0][6]=sbx1[A[6]]; ff[1][6]=f_mult2(ff[0][6]); ff[2][6]=f_mult2(ff[1][6]); ff[3][6]=f_mult2(ff[2][6]);
	ff[0][5]=sbx1[A[5]]; ff[1][5]=f_mult2(ff[0][5]); ff[2][5]=f_mult2(ff[1][5]); ff[3][5]=f_mult2(ff[2][5]); 								
	ff[0][4]=sbx1[A[4]]; ff[1][4]=f_mult2(ff[0][4]); ff[2][4]=f_mult2(ff[1][4]); ff[3][4]=f_mult2(ff[2][4]);								
	ff[0][3]=sbx1[A[3]]; ff[1][3]=f_mult2(ff[0][3]); ff[2][3]=f_mult2(ff[1][3]); ff[3][3]=f_mult2(ff[2][3]);									
	ff[0][2]=sbx1[A[2]]; ff[1][2]=f_mult2(ff[0][2]); ff[2][2]=f_mult2(ff[1][2]); ff[3][2]=f_mult2(ff[2][2]);									
	ff[0][1]=sbx1[A[1]]; ff[1][1]=f_mult2(ff[0][1]); ff[2][1]=f_mult2(ff[1][1]); ff[3][1]=f_mult2(ff[2][1]);									
	ff[0][0]=sbx1[A[0]]; ff[1][0]=f_mult2(ff[0][0]); ff[2][0]=f_mult2(ff[1][0]); ff[3][0]=f_mult2(ff[2][0]);
	temp1[7] = ff[0][7] ^ ff[1][6] ^(ff[1][5] ^ ff[3][5])^(ff[0][4] ^ ff[3][4])^ ff[3][3] ^ ff[0][2] ^ ff[2][1] ^ ff[0][0];
	temp1[6] = ff[0][7] ^ ff[0][6] ^ ff[1][5] ^(ff[1][4] ^ ff[3][4])^(ff[0][3] ^ ff[3][3])^ ff[3][2] ^ ff[0][1] ^ ff[2][0];	
	temp1[5] = ff[2][7] ^ ff[0][6] ^ ff[0][5] ^ ff[1][4] ^(ff[1][3] ^ ff[3][3])^(ff[0][2] ^ ff[3][2])^ ff[3][1] ^ ff[0][0];	
	temp1[4] = ff[0][7] ^ ff[2][6] ^ ff[0][5] ^ ff[0][4] ^ ff[1][3] ^(ff[1][2] ^ ff[3][2])^(ff[0][1] ^ ff[3][1])^ ff[3][0];	
	temp1[3] = ff[3][7] ^ ff[0][6] ^ ff[2][5] ^ ff[0][4] ^ ff[0][3] ^ ff[1][2] ^(ff[1][1] ^ ff[3][1])^(ff[0][0] ^ ff[3][0]);	
	temp1[2] =(ff[0][7] ^ ff[3][7])^ ff[3][6] ^ ff[0][5] ^ ff[2][4] ^ ff[0][3] ^ ff[0][2] ^ ff[1][1] ^(ff[1][0] ^ ff[3][0]);	
	temp1[1] =(ff[1][7] ^ ff[3][7])^(ff[0][6] ^ ff[3][6])^ ff[3][5] ^ ff[0][4] ^ ff[2][3] ^ ff[0][2] ^ ff[0][1] ^ ff[1][0];
	temp1[0] = ff[1][7] ^(ff[1][6] ^ ff[3][6])^(ff[0][5] ^ ff[3][5])^ ff[3][4] ^ ff[0][3] ^ ff[2][2] ^ ff[0][1] ^ ff[0][0];
	B[0] ^= temp1[0];	B[1] ^= temp1[1];	B[2] ^= temp1[2];	B[3] ^= temp1[3];										
	B[4] ^= temp1[4];	B[5] ^= temp1[5];	B[6] ^= temp1[6];	B[7] ^= temp1[7];										
	SSS3(C,temp1,3,5);																									
	SSS5(D,temp1,7,1);																									
	ff[0][7]=sbx1[E[7]]; ff[1][7]=f_mult2(ff[0][7]); ff[2][7]=f_mult2(ff[1][7]); ff[3][7]=f_mult2(ff[2][7]);
	ff[0][6]=sbx1[E[6]]; ff[1][6]=f_mult2(ff[0][6]); ff[2][6]=f_mult2(ff[1][6]); ff[3][6]=f_mult2(ff[2][6]);
	ff[0][5]=sbx1[E[5]]; ff[1][5]=f_mult2(ff[0][5]); ff[2][5]=f_mult2(ff[1][5]); ff[3][5]=f_mult2(ff[2][5]); 								
	ff[0][4]=sbx1[E[4]]; ff[1][4]=f_mult2(ff[0][4]); ff[2][4]=f_mult2(ff[1][4]); ff[3][4]=f_mult2(ff[2][4]);								
	ff[0][3]=sbx1[E[3]]; ff[1][3]=f_mult2(ff[0][3]); ff[2][3]=f_mult2(ff[1][3]); ff[3][3]=f_mult2(ff[2][3]);									
	ff[0][2]=sbx1[E[2]]; ff[1][2]=f_mult2(ff[0][2]); ff[2][2]=f_mult2(ff[1][2]); ff[3][2]=f_mult2(ff[2][2]);									
	ff[0][1]=sbx1[E[1]]; ff[1][1]=f_mult2(ff[0][1]); ff[2][1]=f_mult2(ff[1][1]); ff[3][1]=f_mult2(ff[2][1]);									
	ff[0][0]=sbx1[E[0]]; ff[1][0]=f_mult2(ff[0][0]); ff[2][0]=f_mult2(ff[1][0]); ff[3][0]=f_mult2(ff[2][0]);	
	temp1[7] = ff[0][7] ^ ff[1][6] ^(ff[1][5] ^ ff[3][5])^(ff[0][4] ^ ff[3][4])^ ff[3][3] ^ ff[0][2] ^ ff[2][1] ^ ff[0][0];
	temp1[6] = ff[0][7] ^ ff[0][6] ^ ff[1][5] ^(ff[1][4] ^ ff[3][4])^(ff[0][3] ^ ff[3][3])^ ff[3][2] ^ ff[0][1] ^ ff[2][0];	
	temp1[5] = ff[2][7] ^ ff[0][6] ^ ff[0][5] ^ ff[1][4] ^(ff[1][3] ^ ff[3][3])^(ff[0][2] ^ ff[3][2])^ ff[3][1] ^ ff[0][0];	
	temp1[4] = ff[0][7] ^ ff[2][6] ^ ff[0][5] ^ ff[0][4] ^ ff[1][3] ^(ff[1][2] ^ ff[3][2])^(ff[0][1] ^ ff[3][1])^ ff[3][0];	
	temp1[3] = ff[3][7] ^ ff[0][6] ^ ff[2][5] ^ ff[0][4] ^ ff[0][3] ^ ff[1][2] ^(ff[1][1] ^ ff[3][1])^(ff[0][0] ^ ff[3][0]);	
	temp1[2] =(ff[0][7] ^ ff[3][7])^ ff[3][6] ^ ff[0][5] ^ ff[2][4] ^ ff[0][3] ^ ff[0][2] ^ ff[1][1] ^(ff[1][0] ^ ff[3][0]);	
	temp1[1] =(ff[1][7] ^ ff[3][7])^(ff[0][6] ^ ff[3][6])^ ff[3][5] ^ ff[0][4] ^ ff[2][3] ^ ff[0][2] ^ ff[0][1] ^ ff[1][0];
	temp1[0] = ff[1][7] ^(ff[1][6] ^ ff[3][6])^(ff[0][5] ^ ff[3][5])^ ff[3][4] ^ ff[0][3] ^ ff[2][2] ^ ff[0][1] ^ ff[0][0];
	F[0] ^= temp1[0];	F[1] ^= temp1[1];	F[2] ^= temp1[2];	F[3] ^= temp1[3];										
	F[4] ^= temp1[4];	F[5] ^= temp1[5];	F[6] ^= temp1[6];	F[7] ^= temp1[7];										
	SSS6(G,temp1,3,5);																									
	SSS1(H,temp1,3,5);																									
}	

#define COUNT512(){												\
	      state->counter[ 0]++; if(state->counter[ 0]!=0x00);	\
	else {state->counter[ 1]++; if(state->counter[ 1]!=0x00);	\
	else {state->counter[ 2]++; if(state->counter[ 2]!=0x00);	\
	else {state->counter[ 3]++; if(state->counter[ 3]!=0x00);	\
	else {state->counter[ 4]++; if(state->counter[ 4]!=0x00);	\
	else {state->counter[ 5]++; if(state->counter[ 5]!=0x00);	\
	else {state->counter[ 6]++; if(state->counter[ 6]!=0x00);	\
	else {state->counter[ 7]++; if(state->counter[ 7]!=0x00);	\
	else {state->counter[ 8]++; if(state->counter[ 8]!=0x00);	\
	else {state->counter[ 9]++; if(state->counter[ 9]!=0x00);	\
	else {state->counter[10]++; if(state->counter[10]!=0x00);	\
	else {state->counter[11]++; if(state->counter[11]!=0x00);	\
	else {state->counter[12]++; if(state->counter[12]!=0x00);	\
	else {state->counter[13]++; if(state->counter[13]!=0x00);	\
	else {state->counter[14]++; if(state->counter[14]!=0x00);	\
	else {state->counter[15]++; }}}}}}}}}}}}}}}}				\
		
#define COUNT256(){												\
	      state->counter[0]++; if(state->counter[0]!=0x00);		\
	else {state->counter[1]++; if(state->counter[1]!=0x00);		\
	else {state->counter[2]++; if(state->counter[2]!=0x00);		\
	else {state->counter[3]++; if(state->counter[3]!=0x00);		\
	else {state->counter[4]++; if(state->counter[4]!=0x00);		\
	else {state->counter[5]++; if(state->counter[5]!=0x00);		\
	else {state->counter[6]++; if(state->counter[6]!=0x00);		\
	else {state->counter[7]++; }}}}}}}}							\


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Arirang_Compression256
//
// DESCRIPTION    : Counter addition and compression function of Arirang-224 and ARIRANG-256
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////

void Arirang_Compression256(hashState *state)
{
	BYTE	R[32], W[128], temp1[4];
	BYTE	i;

	// Message Scheduling
	for (i = 0; i < 64; i++)
		W[i] = state->block[i];
	for (; i < 128; i++)
		W[i] = 0;

	temp1[0] = (W[ 36] ^ W[ 44] ^ W[ 52] ^ W[ 60] ^ K256[  0]);	temp1[1] = (W[ 37] ^ W[ 45] ^ W[ 53] ^ W[ 61] ^ K256[  1]);
	temp1[2] = (W[ 38] ^ W[ 46] ^ W[ 54] ^ W[ 62] ^ K256[  2]);	temp1[3] = (W[ 39] ^ W[ 47] ^ W[ 55] ^ W[ 63] ^ K256[  3]);
	SS0((W+ 64),temp1,3,5);
	temp1[0] = (W[ 32] ^ W[ 40] ^ W[ 48] ^ W[ 56] ^ K256[  4]);	temp1[1] = (W[ 33] ^ W[ 41] ^ W[ 49] ^ W[ 57] ^ K256[  5]);
	temp1[2] = (W[ 34] ^ W[ 42] ^ W[ 50] ^ W[ 58] ^ K256[  6]);	temp1[3] = (W[ 35] ^ W[ 43] ^ W[ 51] ^ W[ 59] ^ K256[  7]);
	SS1((W+ 68),temp1,5,3);
	temp1[0] = (W[  4] ^ W[ 12] ^ W[ 20] ^ W[ 28] ^ K256[  8]);	temp1[1] = (W[  5] ^ W[ 13] ^ W[ 21] ^ W[ 29] ^ K256[  9]);
	temp1[2] = (W[  6] ^ W[ 14] ^ W[ 22] ^ W[ 30] ^ K256[ 10]);	temp1[3] = (W[  7] ^ W[ 15] ^ W[ 23] ^ W[ 31] ^ K256[ 11]);
	SS2((W+ 72),temp1,5,3);
	temp1[0] = (W[  0] ^ W[  8] ^ W[ 16] ^ W[ 24] ^ K256[ 12]);	temp1[1] = (W[  1] ^ W[  9] ^ W[ 17] ^ W[ 25] ^ K256[ 13]);
	temp1[2] = (W[  2] ^ W[ 10] ^ W[ 18] ^ W[ 26] ^ K256[ 14]);	temp1[3] = (W[  3] ^ W[ 11] ^ W[ 19] ^ W[ 27] ^ K256[ 15]);
	SS3((W+ 76),temp1,1,7);

	temp1[0] = (W[ 56] ^ W[ 16] ^ W[ 40] ^ W[  0] ^ K256[ 16]);	temp1[1] = (W[ 57] ^ W[ 17] ^ W[ 41] ^ W[  1] ^ K256[ 17]);
	temp1[2] = (W[ 58] ^ W[ 18] ^ W[ 42] ^ W[  2] ^ K256[ 18]);	temp1[3] = (W[ 59] ^ W[ 19] ^ W[ 43] ^ W[  3] ^ K256[ 19]);
	SS0((W+ 80),temp1,3,5);
	temp1[0] = (W[ 44] ^ W[  4] ^ W[ 28] ^ W[ 52] ^ K256[ 20]);	temp1[1] = (W[ 45] ^ W[  5] ^ W[ 29] ^ W[ 53] ^ K256[ 21]);
	temp1[2] = (W[ 46] ^ W[  6] ^ W[ 30] ^ W[ 54] ^ K256[ 22]);	temp1[3] = (W[ 47] ^ W[  7] ^ W[ 31] ^ W[ 55] ^ K256[ 23]);
	SS1((W+ 84),temp1,5,3);
	temp1[0] = (W[ 24] ^ W[ 48] ^ W[  8] ^ W[ 32] ^ K256[ 24]);	temp1[1] = (W[ 25] ^ W[ 49] ^ W[  9] ^ W[ 33] ^ K256[ 25]);
	temp1[2] = (W[ 26] ^ W[ 50] ^ W[ 10] ^ W[ 34] ^ K256[ 26]);	temp1[3] = (W[ 27] ^ W[ 51] ^ W[ 11] ^ W[ 35] ^ K256[ 27]);
	SS2((W+ 88),temp1,5,3);
	temp1[0] = (W[ 12] ^ W[ 36] ^ W[ 60] ^ W[ 20] ^ K256[ 28]);	temp1[1] = (W[ 13] ^ W[ 37] ^ W[ 61] ^ W[ 21] ^ K256[ 29]);
	temp1[2] = (W[ 14] ^ W[ 38] ^ W[ 62] ^ W[ 22] ^ K256[ 30]);	temp1[3] = (W[ 15] ^ W[ 39] ^ W[ 63] ^ W[ 23] ^ K256[ 31]);
	SS3((W+ 92),temp1,1,7);

	temp1[0] = (W[ 52] ^ W[ 60] ^ W[  4] ^ W[ 12] ^ K256[ 32]);	temp1[1] = (W[ 53] ^ W[ 61] ^ W[  5] ^ W[ 13] ^ K256[ 33]);
	temp1[2] = (W[ 54] ^ W[ 62] ^ W[  6] ^ W[ 14] ^ K256[ 34]);	temp1[3] = (W[ 55] ^ W[ 63] ^ W[  7] ^ W[ 15] ^ K256[ 35]);
	SS0((W+ 96),temp1,3,5);
	temp1[0] = (W[ 16] ^ W[ 24] ^ W[ 32] ^ W[ 40] ^ K256[ 36]);	temp1[1] = (W[ 17] ^ W[ 25] ^ W[ 33] ^ W[ 41] ^ K256[ 37]);
	temp1[2] = (W[ 18] ^ W[ 26] ^ W[ 34] ^ W[ 42] ^ K256[ 38]);	temp1[3] = (W[ 19] ^ W[ 27] ^ W[ 35] ^ W[ 43] ^ K256[ 39]);
	SS1((W+100),temp1,5,3);
	temp1[0] = (W[ 20] ^ W[ 28] ^ W[ 36] ^ W[ 44] ^ K256[ 40]);	temp1[1] = (W[ 21] ^ W[ 29] ^ W[ 37] ^ W[ 45] ^ K256[ 41]);
	temp1[2] = (W[ 22] ^ W[ 30] ^ W[ 38] ^ W[ 46] ^ K256[ 42]);	temp1[3] = (W[ 23] ^ W[ 31] ^ W[ 39] ^ W[ 47] ^ K256[ 43]);
	SS2((W+104),temp1,5,3);
	temp1[0] = (W[ 48] ^ W[ 56] ^ W[  0] ^ W[  8] ^ K256[ 44]);	temp1[1] = (W[ 49] ^ W[ 57] ^ W[  1] ^ W[  9] ^ K256[ 45]);
	temp1[2] = (W[ 50] ^ W[ 58] ^ W[  2] ^ W[ 10] ^ K256[ 46]);	temp1[3] = (W[ 51] ^ W[ 59] ^ W[  3] ^ W[ 11] ^ K256[ 47]);
	SS3((W+108),temp1,1,7);

	temp1[0] = (W[ 40] ^ W[  0] ^ W[ 24] ^ W[ 48] ^ K256[ 48]);	temp1[1] = (W[ 41] ^ W[  1] ^ W[ 25] ^ W[ 49] ^ K256[ 49]);
	temp1[2] = (W[ 42] ^ W[  2] ^ W[ 26] ^ W[ 50] ^ K256[ 50]);	temp1[3] = (W[ 43] ^ W[  3] ^ W[ 27] ^ W[ 51] ^ K256[ 51]);
	SS0((W+112),temp1,3,5);
	temp1[0] = (W[ 60] ^ W[ 20] ^ W[ 44] ^ W[  4] ^ K256[ 52]);	temp1[1] = (W[ 61] ^ W[ 21] ^ W[ 45] ^ W[  5] ^ K256[ 53]);
	temp1[2] = (W[ 62] ^ W[ 22] ^ W[ 46] ^ W[  6] ^ K256[ 54]);	temp1[3] = (W[ 63] ^ W[ 23] ^ W[ 47] ^ W[  7] ^ K256[ 55]);
	SS1((W+116),temp1,5,3);
	temp1[0] = (W[  8] ^ W[ 32] ^ W[ 56] ^ W[ 16] ^ K256[ 56]);	temp1[1] = (W[  9] ^ W[ 33] ^ W[ 57] ^ W[ 17] ^ K256[ 57]);
	temp1[2] = (W[ 10] ^ W[ 34] ^ W[ 58] ^ W[ 18] ^ K256[ 58]);	temp1[3] = (W[ 11] ^ W[ 35] ^ W[ 59] ^ W[ 19] ^ K256[ 59]);
	SS2((W+120),temp1,5,3);
	temp1[0] = (W[ 28] ^ W[ 52] ^ W[ 12] ^ W[ 36] ^ K256[ 60]);	temp1[1] = (W[ 29] ^ W[ 53] ^ W[ 13] ^ W[ 37] ^ K256[ 61]);
	temp1[2] = (W[ 30] ^ W[ 54] ^ W[ 14] ^ W[ 38] ^ K256[ 62]);	temp1[3] = (W[ 31] ^ W[ 55] ^ W[ 15] ^ W[ 39] ^ K256[ 63]);
	SS3((W+124),temp1,1,7);

	// Counter Addition
	state->workingvar[ 0] ^= state->counter[7];
	state->workingvar[ 1] ^= state->counter[6];
	state->workingvar[ 2] ^= state->counter[5];
	state->workingvar[ 3] ^= state->counter[4];
	state->workingvar[16] ^= state->counter[3];
	state->workingvar[17] ^= state->counter[2];
	state->workingvar[18] ^= state->counter[1];
	state->workingvar[19] ^= state->counter[0];

	// Initialize Register
	for(i=0;i<32;i++)	R[ i] = state->workingvar[ i];	

	// 1 round                                                                                                                                 
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 64),  (W+ 68));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+  0),  (W+  4));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+  8),  (W+ 12));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 16),  (W+ 20));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 24),  (W+ 28));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 72),  (W+ 76));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 32),  (W+ 36));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 40),  (W+ 44));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 48),  (W+ 52));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+ 56),  (W+ 60));
	                                                                                                                                           
	// 2 round                                                                                                                                 
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 80),  (W+ 84));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 12),  (W+ 24));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 36),  (W+ 48));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 60),  (W+  8));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 20),  (W+ 32));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 88),  (W+ 92));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 44),  (W+ 56));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+  4),  (W+ 16));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 28),  (W+ 40));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 52),  (W+  0));

	// Feedforward_1
	for(i=0;i<32;i++)	R[ i] ^= state->workingvar[(i+16) & 0x1f];
	
	// 3 round                                                                                                                                 
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 96),  (W+100));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 48),  (W+ 20));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 56),  (W+ 28));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+  0),  (W+ 36));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+  8),  (W+ 44));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+104),  (W+108));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 16),  (W+ 52));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+ 24),  (W+ 60));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 32),  (W+  4));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 40),  (W+ 12));
	                                                                                                                                           
	// 4 round                                                                                                                                 
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+112),  (W+116));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+ 28),  (W+  8));
	Step256((R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (W+ 52),  (W+ 32));
	Step256((R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (W+ 12),  (W+ 56));
	Step256((R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (W+ 36),  (W+ 16));
	Step256((R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (R+16),  (W+120),  (W+124));
	Step256((R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (R+12),  (W+ 60),  (W+ 40));
	Step256((R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (R+ 8),  (W+ 20),  (W+  0));
	Step256((R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (R+ 4),  (W+ 44),  (W+ 24));
	Step256((R+ 4),  (R+ 8),  (R+12),  (R+16),  (R+20),  (R+24),  (R+28),  (R+ 0),  (W+  4),  (W+ 48));

	// Feedforward_2
	for(i=0;i<32;i++)	state->workingvar[ i] ^= R[ i]; 
	
	COUNT256();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : ARIRANG_Compression512
//
// DESCRIPTION    : Counter Addition and compression function of ARIRANG-384 and ARIRANG-512
//
// PARAMETERS     : state - a structure that holds the hashState information
//
// RETURNS        : void
//
///////////////////////////////////////////////////////////////////////////////////////////////////
void Arirang_Compression512(hashState *state)
{
	BYTE	R[64], W[256], temp1[8];
	BYTE	*R_temp=R;
	BYTE i;
	
	// Message Scheduling
	for (i = 0; i < 128; i++)	W[i] = state->block[i];
	memset(W+128,0,128);
									
	temp1[0] = (W[ 72] ^ W[ 88] ^ W[104] ^ W[120] ^ K512[  0]);     temp1[1] = (W[ 73] ^ W[ 89] ^ W[105] ^ W[121] ^ K512[  1]);
	temp1[2] = (W[ 74] ^ W[ 90] ^ W[106] ^ W[122] ^ K512[  2]);     temp1[3] = (W[ 75] ^ W[ 91] ^ W[107] ^ W[123] ^ K512[  3]);
	temp1[4] = (W[ 76] ^ W[ 92] ^ W[108] ^ W[124] ^ K512[  4]);     temp1[5] = (W[ 77] ^ W[ 93] ^ W[109] ^ W[125] ^ K512[  5]);
	temp1[6] = (W[ 78] ^ W[ 94] ^ W[110] ^ W[126] ^ K512[  6]);     temp1[7] = (W[ 79] ^ W[ 95] ^ W[111] ^ W[127] ^ K512[  7]);
	SSS1((W+128),temp1,5,3);
	temp1[0] = (W[ 64] ^ W[ 80] ^ W[ 96] ^ W[112] ^ K512[  8]);     temp1[1] = (W[ 65] ^ W[ 81] ^ W[ 97] ^ W[113] ^ K512[  9]);
	temp1[2] = (W[ 66] ^ W[ 82] ^ W[ 98] ^ W[114] ^ K512[ 10]);     temp1[3] = (W[ 67] ^ W[ 83] ^ W[ 99] ^ W[115] ^ K512[ 11]);
	temp1[4] = (W[ 68] ^ W[ 84] ^ W[100] ^ W[116] ^ K512[ 12]);     temp1[5] = (W[ 69] ^ W[ 85] ^ W[101] ^ W[117] ^ K512[ 13]);
	temp1[6] = (W[ 70] ^ W[ 86] ^ W[102] ^ W[118] ^ K512[ 14]);     temp1[7] = (W[ 71] ^ W[ 87] ^ W[103] ^ W[119] ^ K512[ 15]);
	SSS2((W+136),temp1,1,7);
	temp1[0] = (W[  8] ^ W[ 24] ^ W[ 40] ^ W[ 56] ^ K512[ 16]);     temp1[1] = (W[  9] ^ W[ 25] ^ W[ 41] ^ W[ 57] ^ K512[ 17]);
	temp1[2] = (W[ 10] ^ W[ 26] ^ W[ 42] ^ W[ 58] ^ K512[ 18]);     temp1[3] = (W[ 11] ^ W[ 27] ^ W[ 43] ^ W[ 59] ^ K512[ 19]);
	temp1[4] = (W[ 12] ^ W[ 28] ^ W[ 44] ^ W[ 60] ^ K512[ 20]);     temp1[5] = (W[ 13] ^ W[ 29] ^ W[ 45] ^ W[ 61] ^ K512[ 21]);
	temp1[6] = (W[ 14] ^ W[ 30] ^ W[ 46] ^ W[ 62] ^ K512[ 22]);     temp1[7] = (W[ 15] ^ W[ 31] ^ W[ 47] ^ W[ 63] ^ K512[ 23]);
	SSS4((W+144),temp1,3,5);
	temp1[0] = (W[  0] ^ W[ 16] ^ W[ 32] ^ W[ 48] ^ K512[ 24]);     temp1[1] = (W[  1] ^ W[ 17] ^ W[ 33] ^ W[ 49] ^ K512[ 25]);
	temp1[2] = (W[  2] ^ W[ 18] ^ W[ 34] ^ W[ 50] ^ K512[ 26]);     temp1[3] = (W[  3] ^ W[ 19] ^ W[ 35] ^ W[ 51] ^ K512[ 27]);
	temp1[4] = (W[  4] ^ W[ 20] ^ W[ 36] ^ W[ 52] ^ K512[ 28]);     temp1[5] = (W[  5] ^ W[ 21] ^ W[ 37] ^ W[ 53] ^ K512[ 29]);
	temp1[6] = (W[  6] ^ W[ 22] ^ W[ 38] ^ W[ 54] ^ K512[ 30]);     temp1[7] = (W[  7] ^ W[ 23] ^ W[ 39] ^ W[ 55] ^ K512[ 31]);
	SSS7((W+152),temp1,5,3);
	temp1[0] = (W[112] ^ W[ 32] ^ W[ 80] ^ W[  0] ^ K512[ 32]);     temp1[1] = (W[113] ^ W[ 33] ^ W[ 81] ^ W[  1] ^ K512[ 33]);
	temp1[2] = (W[114] ^ W[ 34] ^ W[ 82] ^ W[  2] ^ K512[ 34]);     temp1[3] = (W[115] ^ W[ 35] ^ W[ 83] ^ W[  3] ^ K512[ 35]);
	temp1[4] = (W[116] ^ W[ 36] ^ W[ 84] ^ W[  4] ^ K512[ 36]);     temp1[5] = (W[117] ^ W[ 37] ^ W[ 85] ^ W[  5] ^ K512[ 37]);
	temp1[6] = (W[118] ^ W[ 38] ^ W[ 86] ^ W[  6] ^ K512[ 38]);     temp1[7] = (W[119] ^ W[ 39] ^ W[ 87] ^ W[  7] ^ K512[ 39]);
	SSS1((W+160),temp1,5,3);
	temp1[0] = (W[ 88] ^ W[  8] ^ W[ 56] ^ W[104] ^ K512[ 40]);     temp1[1] = (W[ 89] ^ W[  9] ^ W[ 57] ^ W[105] ^ K512[ 41]);
	temp1[2] = (W[ 90] ^ W[ 10] ^ W[ 58] ^ W[106] ^ K512[ 42]);     temp1[3] = (W[ 91] ^ W[ 11] ^ W[ 59] ^ W[107] ^ K512[ 43]);
	temp1[4] = (W[ 92] ^ W[ 12] ^ W[ 60] ^ W[108] ^ K512[ 44]);     temp1[5] = (W[ 93] ^ W[ 13] ^ W[ 61] ^ W[109] ^ K512[ 45]);
	temp1[6] = (W[ 94] ^ W[ 14] ^ W[ 62] ^ W[110] ^ K512[ 46]);     temp1[7] = (W[ 95] ^ W[ 15] ^ W[ 63] ^ W[111] ^ K512[ 47]);
	SSS2((W+168),temp1,1,7);
	temp1[0] = (W[ 48] ^ W[ 96] ^ W[ 16] ^ W[ 64] ^ K512[ 48]);     temp1[1] = (W[ 49] ^ W[ 97] ^ W[ 17] ^ W[ 65] ^ K512[ 49]);
	temp1[2] = (W[ 50] ^ W[ 98] ^ W[ 18] ^ W[ 66] ^ K512[ 50]);     temp1[3] = (W[ 51] ^ W[ 99] ^ W[ 19] ^ W[ 67] ^ K512[ 51]);
	temp1[4] = (W[ 52] ^ W[100] ^ W[ 20] ^ W[ 68] ^ K512[ 52]);     temp1[5] = (W[ 53] ^ W[101] ^ W[ 21] ^ W[ 69] ^ K512[ 53]);
	temp1[6] = (W[ 54] ^ W[102] ^ W[ 22] ^ W[ 70] ^ K512[ 54]);     temp1[7] = (W[ 55] ^ W[103] ^ W[ 23] ^ W[ 71] ^ K512[ 55]);
	SSS4((W+176),temp1,3,5);
	temp1[0] = (W[ 24] ^ W[ 72] ^ W[120] ^ W[ 40] ^ K512[ 56]);     temp1[1] = (W[ 25] ^ W[ 73] ^ W[121] ^ W[ 41] ^ K512[ 57]);
	temp1[2] = (W[ 26] ^ W[ 74] ^ W[122] ^ W[ 42] ^ K512[ 58]);     temp1[3] = (W[ 27] ^ W[ 75] ^ W[123] ^ W[ 43] ^ K512[ 59]);
	temp1[4] = (W[ 28] ^ W[ 76] ^ W[124] ^ W[ 44] ^ K512[ 60]);     temp1[5] = (W[ 29] ^ W[ 77] ^ W[125] ^ W[ 45] ^ K512[ 61]);
	temp1[6] = (W[ 30] ^ W[ 78] ^ W[126] ^ W[ 46] ^ K512[ 62]);     temp1[7] = (W[ 31] ^ W[ 79] ^ W[127] ^ W[ 47] ^ K512[ 63]);
	SSS7((W+184),temp1,5,3);
	temp1[0] = (W[104] ^ W[120] ^ W[  8] ^ W[ 24] ^ K512[ 64]);     temp1[1] = (W[105] ^ W[121] ^ W[  9] ^ W[ 25] ^ K512[ 65]);
	temp1[2] = (W[106] ^ W[122] ^ W[ 10] ^ W[ 26] ^ K512[ 66]);     temp1[3] = (W[107] ^ W[123] ^ W[ 11] ^ W[ 27] ^ K512[ 67]);
	temp1[4] = (W[108] ^ W[124] ^ W[ 12] ^ W[ 28] ^ K512[ 68]);     temp1[5] = (W[109] ^ W[125] ^ W[ 13] ^ W[ 29] ^ K512[ 69]);
	temp1[6] = (W[110] ^ W[126] ^ W[ 14] ^ W[ 30] ^ K512[ 70]);     temp1[7] = (W[111] ^ W[127] ^ W[ 15] ^ W[ 31] ^ K512[ 71]);
	SSS1((W+192),temp1,5,3);
	temp1[0] = (W[ 32] ^ W[ 48] ^ W[ 64] ^ W[ 80] ^ K512[ 72]);     temp1[1] = (W[ 33] ^ W[ 49] ^ W[ 65] ^ W[ 81] ^ K512[ 73]);
	temp1[2] = (W[ 34] ^ W[ 50] ^ W[ 66] ^ W[ 82] ^ K512[ 74]);     temp1[3] = (W[ 35] ^ W[ 51] ^ W[ 67] ^ W[ 83] ^ K512[ 75]);
	temp1[4] = (W[ 36] ^ W[ 52] ^ W[ 68] ^ W[ 84] ^ K512[ 76]);     temp1[5] = (W[ 37] ^ W[ 53] ^ W[ 69] ^ W[ 85] ^ K512[ 77]);
	temp1[6] = (W[ 38] ^ W[ 54] ^ W[ 70] ^ W[ 86] ^ K512[ 78]);     temp1[7] = (W[ 39] ^ W[ 55] ^ W[ 71] ^ W[ 87] ^ K512[ 79]);
	SSS2((W+200),temp1,1,7);
	temp1[0] = (W[ 40] ^ W[ 56] ^ W[ 72] ^ W[ 88] ^ K512[ 80]);     temp1[1] = (W[ 41] ^ W[ 57] ^ W[ 73] ^ W[ 89] ^ K512[ 81]);
	temp1[2] = (W[ 42] ^ W[ 58] ^ W[ 74] ^ W[ 90] ^ K512[ 82]);     temp1[3] = (W[ 43] ^ W[ 59] ^ W[ 75] ^ W[ 91] ^ K512[ 83]);
	temp1[4] = (W[ 44] ^ W[ 60] ^ W[ 76] ^ W[ 92] ^ K512[ 84]);     temp1[5] = (W[ 45] ^ W[ 61] ^ W[ 77] ^ W[ 93] ^ K512[ 85]);
	temp1[6] = (W[ 46] ^ W[ 62] ^ W[ 78] ^ W[ 94] ^ K512[ 86]);     temp1[7] = (W[ 47] ^ W[ 63] ^ W[ 79] ^ W[ 95] ^ K512[ 87]);
	SSS4((W+208),temp1,3,5);
	temp1[0] = (W[ 96] ^ W[112] ^ W[  0] ^ W[ 16] ^ K512[ 88]);     temp1[1] = (W[ 97] ^ W[113] ^ W[  1] ^ W[ 17] ^ K512[ 89]);
	temp1[2] = (W[ 98] ^ W[114] ^ W[  2] ^ W[ 18] ^ K512[ 90]);     temp1[3] = (W[ 99] ^ W[115] ^ W[  3] ^ W[ 19] ^ K512[ 91]);
	temp1[4] = (W[100] ^ W[116] ^ W[  4] ^ W[ 20] ^ K512[ 92]);     temp1[5] = (W[101] ^ W[117] ^ W[  5] ^ W[ 21] ^ K512[ 93]);
	temp1[6] = (W[102] ^ W[118] ^ W[  6] ^ W[ 22] ^ K512[ 94]);     temp1[7] = (W[103] ^ W[119] ^ W[  7] ^ W[ 23] ^ K512[ 95]);
	SSS7((W+216),temp1,5,3);
	temp1[0] = (W[ 80] ^ W[  0] ^ W[ 48] ^ W[ 96] ^ K512[ 96]);     temp1[1] = (W[ 81] ^ W[  1] ^ W[ 49] ^ W[ 97] ^ K512[ 97]);
	temp1[2] = (W[ 82] ^ W[  2] ^ W[ 50] ^ W[ 98] ^ K512[ 98]);     temp1[3] = (W[ 83] ^ W[  3] ^ W[ 51] ^ W[ 99] ^ K512[ 99]);
	temp1[4] = (W[ 84] ^ W[  4] ^ W[ 52] ^ W[100] ^ K512[100]);     temp1[5] = (W[ 85] ^ W[  5] ^ W[ 53] ^ W[101] ^ K512[101]);
	temp1[6] = (W[ 86] ^ W[  6] ^ W[ 54] ^ W[102] ^ K512[102]);     temp1[7] = (W[ 87] ^ W[  7] ^ W[ 55] ^ W[103] ^ K512[103]);
	SSS1((W+224),temp1,5,3);
	temp1[0] = (W[120] ^ W[ 40] ^ W[ 88] ^ W[  8] ^ K512[104]);     temp1[1] = (W[121] ^ W[ 41] ^ W[ 89] ^ W[  9] ^ K512[105]);
	temp1[2] = (W[122] ^ W[ 42] ^ W[ 90] ^ W[ 10] ^ K512[106]);     temp1[3] = (W[123] ^ W[ 43] ^ W[ 91] ^ W[ 11] ^ K512[107]);
	temp1[4] = (W[124] ^ W[ 44] ^ W[ 92] ^ W[ 12] ^ K512[108]);     temp1[5] = (W[125] ^ W[ 45] ^ W[ 93] ^ W[ 13] ^ K512[109]);
	temp1[6] = (W[126] ^ W[ 46] ^ W[ 94] ^ W[ 14] ^ K512[110]);     temp1[7] = (W[127] ^ W[ 47] ^ W[ 95] ^ W[ 15] ^ K512[111]);
	SSS2((W+232),temp1,1,7);
	temp1[0] = (W[ 16] ^ W[ 64] ^ W[112] ^ W[ 32] ^ K512[112]);     temp1[1] = (W[ 17] ^ W[ 65] ^ W[113] ^ W[ 33] ^ K512[113]);
	temp1[2] = (W[ 18] ^ W[ 66] ^ W[114] ^ W[ 34] ^ K512[114]);     temp1[3] = (W[ 19] ^ W[ 67] ^ W[115] ^ W[ 35] ^ K512[115]);
	temp1[4] = (W[ 20] ^ W[ 68] ^ W[116] ^ W[ 36] ^ K512[116]);     temp1[5] = (W[ 21] ^ W[ 69] ^ W[117] ^ W[ 37] ^ K512[117]);
	temp1[6] = (W[ 22] ^ W[ 70] ^ W[118] ^ W[ 38] ^ K512[118]);     temp1[7] = (W[ 23] ^ W[ 71] ^ W[119] ^ W[ 39] ^ K512[119]);
	SSS4((W+240),temp1,3,5);
	temp1[0] = (W[ 56] ^ W[104] ^ W[ 24] ^ W[ 72] ^ K512[120]);     temp1[1] = (W[ 57] ^ W[105] ^ W[ 25] ^ W[ 73] ^ K512[121]);
	temp1[2] = (W[ 58] ^ W[106] ^ W[ 26] ^ W[ 74] ^ K512[122]);     temp1[3] = (W[ 59] ^ W[107] ^ W[ 27] ^ W[ 75] ^ K512[123]);
	temp1[4] = (W[ 60] ^ W[108] ^ W[ 28] ^ W[ 76] ^ K512[124]);     temp1[5] = (W[ 61] ^ W[109] ^ W[ 29] ^ W[ 77] ^ K512[125]);
	temp1[6] = (W[ 62] ^ W[110] ^ W[ 30] ^ W[ 78] ^ K512[126]);     temp1[7] = (W[ 63] ^ W[111] ^ W[ 31] ^ W[ 79] ^ K512[127]);
	SSS7((W+248),temp1,5,3);																														

	// Counter Addition
	for(i=0;i<8;i++) state->workingvar[ i] ^= state->counter[15-i];
	for(i=32;i<40;i++) state->workingvar[ i] ^= state->counter[39-i];
	
	// Initialize Register
	for(i=0;i<64;i++) R[ i] = state->workingvar[ i];	
	
	// 1 Round                                                                                      
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+128), (W+136));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+  0), (W+  8));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 16), (W+ 24));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 32), (W+ 40));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 48), (W+ 56));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+144), (W+152));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 64), (W+ 72));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+ 80), (W+ 88));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 96), (W+104));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+112), (W+120));
                                                                                                
	// 2 round                                                                                    
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+160), (W+168));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 24), (W+ 48));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 72), (W+ 96));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+120), (W+ 16));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 40), (W+ 64));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+176), (W+184));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 88), (W+112));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+  8), (W+ 32));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 56), (W+ 80));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+104), (W+  0));



	// Feedforward_1
	for(i=0;i<64;i++) R[ i] ^= state->workingvar[(32+i)&0x3f];
	                                                                                   
	// 3 Round                                                                                    
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+192), (W+200));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 96), (W+ 40));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+112), (W+ 56));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+  0), (W+ 72));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+ 16), (W+ 88));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+208), (W+216));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 32), (W+104));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+ 48), (W+120));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+ 64), (W+  8));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 80), (W+ 24));
	                                                                                              
	// 4 round                                                                                    
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+224), (W+232));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+ 56), (W+ 16));
	Step512((R+ 0), (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56), (W+104), (W+ 64));
	Step512((R+56), (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40),  (R+48), (W+ 24), (W+112));
	Step512((R+48), (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32),  (R+40), (W+ 72), (W+ 32));
	Step512((R+40), (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24),  (R+32), (W+240), (W+248));
	Step512((R+32), (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16),  (R+24), (W+120), (W+ 80));
	Step512((R+24), (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8),  (R+16), (W+ 40), (W+  0));
	Step512((R+16), (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0),  (R+ 8), (W+ 88), (W+ 48));
	Step512((R+ 8), (R+16),  (R+24),  (R+32),  (R+40),  (R+48),  (R+56),  (R+ 0), (W+  8), (W+ 96));
	

	// Feedforward_2
	for(i=0;i<64;i++) state->workingvar[ i] ^= R[ i]; 
	
	COUNT512();														
}

///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Init
//
// DESCRIPTION    : Initialize a hashState with the intended hash length of particular instantiation.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  hashbytelen - an integer value that indicates the length of the hash output in bytes
//
// RETURNS        : SUCCESS - on success
//                  BAD_HASHLEN - hashbytelen is invalid
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Init(hashState *state, BYTE hashbytelen)
{
	if ((hashbytelen != 28) && (hashbytelen != 32) && (hashbytelen != 48) && (hashbytelen != 64))
		return BAD_HASHLEN;

	// Setting the Hash Length
	state->hashbytelen = hashbytelen;
	
	// Setting the Counter Values
	memset(state->counter,0,16);
	
	// Initialize the Data Length
	memset(state->count,0,16);

	// Setting the Initial Hash Value
	if(state->hashbytelen == 28){
		state->workingvar[ 0] = 0xcb; state->workingvar[ 1] = 0xbb; state->workingvar[ 2] = 0x9d; state->workingvar[ 3] = 0x5d;
		state->workingvar[ 4] = 0x62; state->workingvar[ 5] = 0x9a; state->workingvar[ 6] = 0x29; state->workingvar[ 7] = 0x2a;
		state->workingvar[ 8] = 0x91; state->workingvar[ 9] = 0x59; state->workingvar[10] = 0x01; state->workingvar[11] = 0x5a;
		state->workingvar[12] = 0x15; state->workingvar[13] = 0x2f; state->workingvar[14] = 0xec; state->workingvar[15] = 0xd8;
		state->workingvar[16] = 0x67; state->workingvar[17] = 0x33; state->workingvar[18] = 0x26; state->workingvar[19] = 0x67;
		state->workingvar[20] = 0x8e; state->workingvar[21] = 0xb4; state->workingvar[22] = 0x4a; state->workingvar[23] = 0x87;
		state->workingvar[24] = 0xdb; state->workingvar[25] = 0x0c; state->workingvar[26] = 0x2e; state->workingvar[27] = 0x0d;
		state->workingvar[28] = 0x47; state->workingvar[29] = 0xb5; state->workingvar[30] = 0x48; state->workingvar[31] = 0x1d;
		state->blocklen=ARIRANG256_BLOCK_LEN;
	}
		
	else if(state->hashbytelen == 32){
		state->workingvar[ 0] = 0x6a; state->workingvar[ 1] = 0x09; state->workingvar[ 2] = 0xe6; state->workingvar[ 3] = 0x67;
		state->workingvar[ 4] = 0xbb; state->workingvar[ 5] = 0x67; state->workingvar[ 6] = 0xae; state->workingvar[ 7] = 0x85;
		state->workingvar[ 8] = 0x3c; state->workingvar[ 9] = 0x6e; state->workingvar[10] = 0xf3; state->workingvar[11] = 0x72;
		state->workingvar[12] = 0xa5; state->workingvar[13] = 0x4f; state->workingvar[14] = 0xf5; state->workingvar[15] = 0x3a;
		state->workingvar[16] = 0x51; state->workingvar[17] = 0x0e; state->workingvar[18] = 0x52; state->workingvar[19] = 0x7f;
		state->workingvar[20] = 0x9b; state->workingvar[21] = 0x05; state->workingvar[22] = 0x68; state->workingvar[23] = 0x8c;
		state->workingvar[24] = 0x1f; state->workingvar[25] = 0x83; state->workingvar[26] = 0xd9; state->workingvar[27] = 0xab;
		state->workingvar[28] = 0x5b; state->workingvar[29] = 0xe0; state->workingvar[30] = 0xcd; state->workingvar[31] = 0x19;
		state->blocklen=ARIRANG256_BLOCK_LEN;
	
	}
	else if(hashbytelen == 48){	                                                                                       
		state->workingvar[ 0]=0xcb; state->workingvar[ 1]=0xbb; state->workingvar[ 2]=0x9d; state->workingvar[ 3]=0x5d;
		state->workingvar[ 4]=0xc1; state->workingvar[ 5]=0x05; state->workingvar[ 6]=0x9e; state->workingvar[ 7]=0xd8;
		state->workingvar[ 8]=0x62; state->workingvar[ 9]=0x9a; state->workingvar[10]=0x29; state->workingvar[11]=0x2a;
		state->workingvar[12]=0x36; state->workingvar[13]=0x7c; state->workingvar[14]=0xd5; state->workingvar[15]=0x07;
		state->workingvar[16]=0x91; state->workingvar[17]=0x59; state->workingvar[18]=0x01; state->workingvar[19]=0x5a;
		state->workingvar[20]=0x30; state->workingvar[21]=0x70; state->workingvar[22]=0xdd; state->workingvar[23]=0x17;
		state->workingvar[24]=0x15; state->workingvar[25]=0x2f; state->workingvar[26]=0xec; state->workingvar[27]=0xd8;
		state->workingvar[28]=0xf7; state->workingvar[29]=0x0e; state->workingvar[30]=0x59; state->workingvar[31]=0x39;
		state->workingvar[32]=0x67; state->workingvar[33]=0x33; state->workingvar[34]=0x26; state->workingvar[35]=0x67;
		state->workingvar[36]=0xff; state->workingvar[37]=0xc0; state->workingvar[38]=0x0b; state->workingvar[39]=0x31;
		state->workingvar[40]=0x8e; state->workingvar[41]=0xb4; state->workingvar[42]=0x4a; state->workingvar[43]=0x87;
		state->workingvar[44]=0x68; state->workingvar[45]=0x58; state->workingvar[46]=0x15; state->workingvar[47]=0x11;
		state->workingvar[48]=0xdb; state->workingvar[49]=0x0c; state->workingvar[50]=0x2e; state->workingvar[51]=0x0d;
		state->workingvar[52]=0x64; state->workingvar[53]=0xf9; state->workingvar[54]=0x8f; state->workingvar[55]=0xa7;
		state->workingvar[56]=0x47; state->workingvar[57]=0xb5; state->workingvar[58]=0x48; state->workingvar[59]=0x1d;
		state->workingvar[60]=0xbe; state->workingvar[61]=0xfa; state->workingvar[62]=0x4f; state->workingvar[63]=0xa4;
		state->blocklen=ARIRANG512_BLOCK_LEN;	                                                                         
	}													                                                                                       
	else if(hashbytelen == 64){									                                                                     
		state->workingvar[ 0]=0x6a; state->workingvar[ 1]=0x09; state->workingvar[ 2]=0xe6; state->workingvar[ 3]=0x67;
		state->workingvar[ 4]=0xf3; state->workingvar[ 5]=0xbc; state->workingvar[ 6]=0xc9; state->workingvar[ 7]=0x08;
		state->workingvar[ 8]=0xbb; state->workingvar[ 9]=0x67; state->workingvar[10]=0xae; state->workingvar[11]=0x85;
		state->workingvar[12]=0x84; state->workingvar[13]=0xca; state->workingvar[14]=0xa7; state->workingvar[15]=0x3b;
		state->workingvar[16]=0x3c; state->workingvar[17]=0x6e; state->workingvar[18]=0xf3; state->workingvar[19]=0x72;
		state->workingvar[20]=0xfe; state->workingvar[21]=0x94; state->workingvar[22]=0xf8; state->workingvar[23]=0x2b;
		state->workingvar[24]=0xa5; state->workingvar[25]=0x4f; state->workingvar[26]=0xf5; state->workingvar[27]=0x3a;
		state->workingvar[28]=0x5f; state->workingvar[29]=0x1d; state->workingvar[30]=0x36; state->workingvar[31]=0xf1;
		state->workingvar[32]=0x51; state->workingvar[33]=0x0e; state->workingvar[34]=0x52; state->workingvar[35]=0x7f;
		state->workingvar[36]=0xad; state->workingvar[37]=0xe6; state->workingvar[38]=0x82; state->workingvar[39]=0xd1;
		state->workingvar[40]=0x9b; state->workingvar[41]=0x05; state->workingvar[42]=0x68; state->workingvar[43]=0x8c;
		state->workingvar[44]=0x2b; state->workingvar[45]=0x3e; state->workingvar[46]=0x6c; state->workingvar[47]=0x1f;
		state->workingvar[48]=0x1f; state->workingvar[49]=0x83; state->workingvar[50]=0xd9; state->workingvar[51]=0xab;
		state->workingvar[52]=0xfb; state->workingvar[53]=0x41; state->workingvar[54]=0xbd; state->workingvar[55]=0x6b;
		state->workingvar[56]=0x5b; state->workingvar[57]=0xe0; state->workingvar[58]=0xcd; state->workingvar[59]=0x19;
		state->workingvar[60]=0x13; state->workingvar[61]=0x7e; state->workingvar[62]=0x21; state->workingvar[63]=0x79;
		state->blocklen=ARIRANG512_BLOCK_LEN;		                                                                       
	}                                                                                                                

	
	return SUCCESS;

}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Update
//
// DESCRIPTION    : Process the supplied data.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  data - the data to be hashed
//                  Databitlen - the length, in bits, of the data to be hashed
//
// RETURNS        : SUCCESS - on success
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Update(hashState *state, const BitSequence *data, BYTE* databitlen)
{
	BYTE RemainedLen, PartLen, temp;
	BYTE databytelen;

	
	databytelen = ( databitlen[0]+7 >> 3) + (databitlen[1] << 5 );
	
	
	// Compute the number of hashed bytes mod ARIRANG_BLOCK_LEN
	RemainedLen = ((state->count[0] >> 3) + (state->count[1] << 5)) % state->blocklen;

	// Compute the number of bytes that can be filled up
	PartLen = state->blocklen - RemainedLen;

	// Update count (number of toatl data bits)
	temp = state->count[0] + (databytelen << 3);
	if(temp < state->count[0] ){
		  state->count[ 1]++; if(state->count[ 1] != 0x00);
	else {state->count[ 2]++; if(state->count[ 2] != 0x00); 
	else {state->count[ 3]++; if(state->count[ 3] != 0x00);
	else {state->count[ 4]++; if(state->count[ 4] != 0x00);
	else {state->count[ 5]++; if(state->count[ 5] != 0x00);
	else {state->count[ 6]++; if(state->count[ 6] != 0x00);
	else {state->count[ 7]++; if(state->count[ 7] != 0x00); 
	else {state->count[ 8]++; if(state->count[ 8] != 0x00);
	else {state->count[ 9]++; if(state->count[ 9] != 0x00);
	else {state->count[10]++; if(state->count[10] != 0x00);
	else {state->count[11]++; if(state->count[11] != 0x00);
	else {state->count[12]++; if(state->count[12] != 0x00); 
	else {state->count[13]++; if(state->count[13] != 0x00);
	else {state->count[14]++; if(state->count[14] != 0x00);
	else {state->count[15]++;}}}}}}}}}}}}}}}
	state->count[0]=temp;

	temp = state->count[1] + databitlen[1];
	if(temp < state->count[1] ){
		  state->count[ 2]++; if(state->count[ 2] != 0x00);
	else {state->count[ 3]++; if(state->count[ 3] != 0x00); 
	else {state->count[ 4]++; if(state->count[ 4] != 0x00);
	else {state->count[ 5]++; if(state->count[ 5] != 0x00);
	else {state->count[ 6]++; if(state->count[ 6] != 0x00);
	else {state->count[ 7]++; if(state->count[ 7] != 0x00); 
	else {state->count[ 8]++; if(state->count[ 8] != 0x00);
	else {state->count[ 9]++; if(state->count[ 9] != 0x00);
	else {state->count[10]++; if(state->count[10] != 0x00);
	else {state->count[11]++; if(state->count[11] != 0x00);
	else {state->count[12]++; if(state->count[12] != 0x00); 
	else {state->count[13]++; if(state->count[13] != 0x00);
	else {state->count[14]++; if(state->count[14] != 0x00);
	else {state->count[15]++;}}}}}}}}}}}}}}
	state->count[1]=temp;
	
	if (((databytelen >= PartLen) && (databitlen[0] == 0)) || ((databytelen == PartLen) && (state->remainderbit == 0)) ) {
		memcpy(state->block + RemainedLen, data, PartLen);
		if(state->hashbytelen < 33) Arirang_Compression256(state);
		else Arirang_Compression512(state);

		data += PartLen;
		databytelen -= PartLen;
		RemainedLen = 0;

		while( (databytelen > state->blocklen) || ((databytelen == state->blocklen) && (state->remainderbit == 0)) ) {
			memcpy((BYTE *)state->block, data, state->blocklen);
			if(state->hashbytelen < 33) Arirang_Compression256(state);
			else Arirang_Compression512(state);

			data += state->blocklen;
			databytelen -= state->blocklen;
		}
	}

	//	Buffer remaining input
	memcpy((BYTE *)state->block + RemainedLen, data, databytelen);

	return SUCCESS;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION NAME  : Final
//
// DESCRIPTION    : Perform any post processing and ouput filtering required and return the final 
//                  hash value.
//
// PARAMETERS     : state - a structure that holds the hashState information
//                  hashval - the storage for the final hash value to be returned
//
// RETURNS        : SUCCESS - on success
//
///////////////////////////////////////////////////////////////////////////////////////////////////

HashReturn Final(hashState *state, BYTE *hashval)
{
	BYTE i, dwIndex, count[16];
	BYTE temp = (state->blocklen >> 3);

	// Padding the message
	if(state->remainderbit){
		count[ 0] = state->count[ 0] + state->remainderbit - 8; count[ 1] = state->count[ 1]; count[ 2] = state->count[ 2]; count[ 3] = state->count[ 3];
		count[ 4] = state->count[ 4]; count[ 5] = state->count[ 5]; count[ 6] = state->count[ 6]; count[ 7] = state->count[ 7];
		count[ 8] = state->count[ 8]; count[ 9] = state->count[ 9]; count[10] = state->count[10]; count[11] = state->count[11];
		count[12] = state->count[12]; count[13] = state->count[13]; count[14] = state->count[14]; count[15] = state->count[15];

		dwIndex = ((count[0] >> 3) + (count[1] << 5)) % state->blocklen;
		state->block[dwIndex] &= 0xff-(1<<(8-state->remainderbit))+1;		
		state->block[dwIndex++] ^= 0x80>>(state->remainderbit);
	}
	else{
		count[ 0] = state->count[ 0]; count[ 1] = state->count[ 1]; count[ 2] = state->count[ 2]; count[ 3] = state->count[ 3];
		count[ 4] = state->count[ 4]; count[ 5] = state->count[ 5]; count[ 6] = state->count[ 6]; count[ 7] = state->count[ 7];
		count[ 8] = state->count[ 8]; count[ 9] = state->count[ 9]; count[10] = state->count[10]; count[11] = state->count[11];
		count[12] = state->count[12]; count[13] = state->count[13]; count[14] = state->count[14]; count[15] = state->count[15];
		dwIndex = ((count[0] >> 3) + (count[1] << 5)) % state->blocklen;		
		state->block[dwIndex++] = 0x80;	
	}
		
	if (dwIndex > (state->blocklen - temp)){
		memset((BYTE *)state->block + dwIndex, 0, (BYTE)(state->blocklen - dwIndex));
		if(state->hashbytelen < 33) Arirang_Compression256(state);
		else Arirang_Compression512(state);

		memset((BYTE *)state->block, 0, (BYTE)state->blocklen - temp);
	}
	else
		memset((BYTE *)state->block + dwIndex, 0, (BYTE)(state->blocklen - dwIndex - temp));

	// Fixed counter value for the last message block
	if(state->hashbytelen > 33){
		state->block[state->blocklen-16] = count[15]; state->block[state->blocklen-15] = count[14];	state->block[state->blocklen-14] = count[13]; state->block[state->blocklen-13] = count[12];
		state->block[state->blocklen-12] = count[11]; state->block[state->blocklen-11] = count[10];	state->block[state->blocklen-10] = count[ 9]; state->block[state->blocklen- 9] = count[ 8];
		state->block[state->blocklen- 8] = count[ 7]; state->block[state->blocklen- 7] = count[ 6];	state->block[state->blocklen- 6] = count[ 5]; state->block[state->blocklen- 5] = count[ 4];
		state->block[state->blocklen- 4] = count[ 3]; state->block[state->blocklen- 3] = count[ 2];	state->block[state->blocklen- 2] = count[ 1]; state->block[state->blocklen- 1] = count[ 0];
		state->counter[15]=0xb7; state->counter[14]=0xe1; state->counter[13]=0x51; state->counter[12]=0x62;	
		state->counter[11]=0x8a; state->counter[10]=0xed; state->counter[ 9]=0x2a; state->counter[ 8]=0x6a;
		state->counter[ 7]=0xbf; state->counter[ 6]=0x71; state->counter[ 5]=0x58; state->counter[ 4]=0x80;	
		state->counter[ 3]=0x9c; state->counter[ 2]=0xf4; state->counter[ 1]=0xf3; state->counter[ 0]=0xc7;
	}
	else{
		state->block[state->blocklen-8] = count[7];	state->block[state->blocklen-7] = count[6];	state->block[state->blocklen-6] = count[5];	state->block[state->blocklen-5] = count[4];
		state->block[state->blocklen-4] = count[3];	state->block[state->blocklen-3] = count[2];	state->block[state->blocklen-2] = count[1];	state->block[state->blocklen-1] = count[0];
		state->counter[7]=0xb7; state->counter[6]=0xe1; state->counter[5]=0x51; state->counter[4]=0x62;	
		state->counter[3]=0x8a; state->counter[2]=0xed; state->counter[1]=0x2a; state->counter[0]=0x6a;	
	}
	
	if(state->hashbytelen < 33) Arirang_Compression256(state);
	else Arirang_Compression512(state);

	for (i = 0; i < state->hashbytelen ; i++)	hashval[i]=state->workingvar[i]; 

	return SUCCESS;
}

HashReturn Hash(BYTE hashbytelen, const BitSequence *data, BYTE *databitlen, BitSequence *hashval){

	hashState State;
	BitSequence *UpdatedData;	
	BYTE UpdatedDataLengthbit[2]={0,1};
	BYTE i[15]={0,};
	
	UpdatedData=data;
	State.remainderbit = databitlen[0] & 7;

	Init(&State, hashbytelen);
	while((i[ 0] < databitlen[ 1]) || (i[ 1] < databitlen[ 2]) || (i[ 2] < databitlen[ 3]) || (i[ 3] < databitlen[ 4]) ||
		  (i[ 4] < databitlen[ 5]) || (i[ 5] < databitlen[ 6]) || (i[ 6] < databitlen[ 7]) || (i[ 7] < databitlen[ 8]) ||
		  (i[ 8] < databitlen[ 9]) || (i[ 9] < databitlen[10]) || (i[10] < databitlen[11]) || (i[11] < databitlen[12]) ||
		  (i[12] < databitlen[13]) || (i[13] < databitlen[14]) || (i[14] < databitlen[15]) ){
		Update(&State, UpdatedData, UpdatedDataLengthbit);
		UpdatedData+=0x20;
		i[0]++;  if((i[0] != 0x00));
		else {i[ 1]++;if(i[ 1] != 0x00);
		else {i[ 2]++;if(i[ 2] != 0x00);
		else {i[ 3]++;if(i[ 3] != 0x00);
		else {i[ 4]++;if(i[ 4] != 0x00);
		else {i[ 5]++;if(i[ 5] != 0x00);
		else {i[ 6]++;if(i[ 6] != 0x00);
		else {i[ 7]++;if(i[ 7] != 0x00);
		else {i[ 8]++;if(i[ 8] != 0x00);
		else {i[ 9]++;if(i[ 9] != 0x00);
		else {i[10]++;if(i[10] != 0x00);
		else {i[11]++;if(i[11] != 0x00);
		else {i[12]++;if(i[12] != 0x00);
		else {i[13]++;if(i[13] != 0x00);
		else {i[14]++;if(i[14] != 0x00);
		else {i[15]++;}}}}}}}}}}}}}}}
	}
	UpdatedDataLengthbit[1] = 0;
	UpdatedDataLengthbit[0] = databitlen[0];
	Update(&State, UpdatedData, UpdatedDataLengthbit);

	Final(&State, hashval);
	return SUCCESS;
	
}